# Description:
# This code solves the 1D Poisson problem
# u''(x) = f 
# where f is a source function on the domain (a,b)
# subject to the boundary conditions
# u(a) = alpha and u(b) = beta
# Reference: http://www-m16.ma.tum.de/foswiki/pub/M16/Allgemeines/StefanPossanner/Poisson1D_FD.html
# Author: Nicholas A. Danes, PhD
# Colorado School of Mines
# Cyber Infrastructure & Advanced Research Computing
# Date: 02/09/2021

import numpy as np
import matplotlib.pyplot as plt

# setup the domain size
a = 0.0
b = 1.0

# define the Dirichlet boundary points 
alpha = 0.0
beta = 0.0

# length of domain
L = b-a

# choose the number of interior grid points ("cells")
N = 100

# define grid spacing
h = L/float(N)

# use a linear space to define a vector of grid points
x = np.linspace(a,b,N+1)

# define a function for our exact solution
# this assumes the source function equals 1
def u_exact_eval(x):
	return (-b**2 * x + a**2 * (-b + x) + b * (x**2 - 2.0*alpha) + 2.0 * x * (alpha - beta) + a * (b**2 - x**2 + 2.0 * beta))/(2.0 * (a - b))

# setup column vector to store exact solution
u_exact = np.zeros( N-1 )
u_exact[:] = u_exact_eval(x[1:N])

# define f, our source term
def f_eval(x):
	return 1

# define the source vector
f = np.zeros(N-1)

# evaluate it using the f_eval function on (a,b)
f[:] = f_eval(x[:])

# since the Dirichlet boundary points are known,
# move the points onto the right hand side of our discrete
# equation
f[0] += alpha/h**2
f[-1] += beta/h**2

# discrete Laplacian
A = np.zeros((N-1, N-1))

## main diagonal for discrete Laplacian 
d_main = np.diag( 2.0*np.ones(N-1) )
d_upper = np.diag( -1.0*np.ones(N-2), +1 )
d_lower = np.diag( -1.0*np.ones(N-2), -1 )

# define the discrete Laplacian and scale by the grid size
# to define the finite difference approximation 
A = d_main + d_upper + d_lower
A *= 1.0/h**2 

# let's solve the system using linalg.solve
u = np.linalg.solve(A, f)

# let's plot the solutions
plt.plot( x[1:N], u_exact, label="Exact Solution" )
plt.plot( x[1:N], u, linestyle='--', label="Numerical Solution" )
plt.legend()
plt.savefig("1D_poisson.png", dpi=200)

# save the solutions to a text file
np.savetxt("poisson_U.txt", u)
np.savetxt("poisson_Uexact.txt", u_exact)
np.savetxt("poisson_X.txt", x)
